# Copyright (c) 2020 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: workspace.proto

import sys
_b = sys.version_info[0] < 3 and (lambda x: x) or (
    lambda x: x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()

DESCRIPTOR = _descriptor.FileDescriptor(
    name='workspace.proto',
    package='suitebackend',
    syntax='proto3',
    serialized_options=None,
    serialized_pb=_b(
        '\n\x0fworkspace.proto\x12\x0csuitebackend\"b\n\x07\x44\x61taset\x12\n\n\x02id\x18\x01 \x01(\t\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x0c\n\x04\x64\x65sc\x18\x03 \x01(\t\x12\x0c\n\x04type\x18\x04 \x01(\t\x12\x0c\n\x04path\x18\x05 \x01(\t\x12\x13\n\x0b\x63reate_time\x18\x06 \x01(\t\"o\n\x07Project\x12\n\n\x02id\x18\x01 \x01(\t\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x0c\n\x04\x64\x65sc\x18\x03 \x01(\t\x12\x0c\n\x04type\x18\x04 \x01(\t\x12\x0b\n\x03\x64id\x18\x05 \x01(\t\x12\x0c\n\x04path\x18\x06 \x01(\t\x12\x13\n\x0b\x63reate_time\x18\x07 \x01(\t\"q\n\x04Task\x12\n\n\x02id\x18\x01 \x01(\t\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x0c\n\x04\x64\x65sc\x18\x03 \x01(\t\x12\x0b\n\x03pid\x18\x04 \x01(\t\x12\x0c\n\x04path\x18\x05 \x01(\t\x12\x13\n\x0b\x63reate_time\x18\x06 \x01(\t\x12\x11\n\tparent_id\x18\x07 \x01(\t\"\x85\x01\n\x0fPretrainedModel\x12\n\n\x02id\x18\x01 \x01(\t\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\r\n\x05model\x18\x03 \x01(\t\x12\x0c\n\x04type\x18\x04 \x01(\t\x12\x0b\n\x03pid\x18\x05 \x01(\t\x12\x0b\n\x03tid\x18\x06 \x01(\t\x12\x13\n\x0b\x63reate_time\x18\x07 \x01(\t\x12\x0c\n\x04path\x18\x08 \x01(\t\"\x9a\x01\n\rExportedModel\x12\n\n\x02id\x18\x01 \x01(\t\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\r\n\x05model\x18\x03 \x01(\t\x12\x0c\n\x04type\x18\x04 \x01(\t\x12\x0b\n\x03pid\x18\x05 \x01(\t\x12\x0b\n\x03tid\x18\x06 \x01(\t\x12\x13\n\x0b\x63reate_time\x18\x07 \x01(\t\x12\x0c\n\x04path\x18\x08 \x01(\t\x12\x15\n\rexported_type\x18\t \x01(\x05\"\xf8\x06\n\tWorkspace\x12\x0f\n\x07version\x18\x01 \x01(\t\x12\x0c\n\x04path\x18\x02 \x01(\t\x12\x37\n\x08\x64\x61tasets\x18\x03 \x03(\x0b\x32%.suitebackend.Workspace.DatasetsEntry\x12\x37\n\x08projects\x18\x04 \x03(\x0b\x32%.suitebackend.Workspace.ProjectsEntry\x12\x31\n\x05tasks\x18\x05 \x03(\x0b\x32\".suitebackend.Workspace.TasksEntry\x12\x16\n\x0emax_dataset_id\x18\x06 \x01(\x05\x12\x16\n\x0emax_project_id\x18\x07 \x01(\x05\x12\x13\n\x0bmax_task_id\x18\x08 \x01(\x05\x12\x14\n\x0c\x63urrent_time\x18\t \x01(\t\x12\x1f\n\x17max_pretrained_model_id\x18\n \x01(\x05\x12H\n\x11pretrained_models\x18\x0b \x03(\x0b\x32-.suitebackend.Workspace.PretrainedModelsEntry\x12\x1d\n\x15max_exported_model_id\x18\x0c \x01(\x05\x12\x44\n\x0f\x65xported_models\x18\r \x03(\x0b\x32+.suitebackend.Workspace.ExportedModelsEntry\x1a\x46\n\rDatasetsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12$\n\x05value\x18\x02 \x01(\x0b\x32\x15.suitebackend.Dataset:\x02\x38\x01\x1a\x46\n\rProjectsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12$\n\x05value\x18\x02 \x01(\x0b\x32\x15.suitebackend.Project:\x02\x38\x01\x1a@\n\nTasksEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12!\n\x05value\x18\x02 \x01(\x0b\x32\x12.suitebackend.Task:\x02\x38\x01\x1aV\n\x15PretrainedModelsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12,\n\x05value\x18\x02 \x01(\x0b\x32\x1d.suitebackend.PretrainedModel:\x02\x38\x01\x1aR\n\x13\x45xportedModelsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12*\n\x05value\x18\x02 \x01(\x0b\x32\x1b.suitebackend.ExportedModel:\x02\x38\x01\x62\x06proto3'
    ))

_DATASET = _descriptor.Descriptor(
    name='Dataset',
    full_name='suitebackend.Dataset',
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name='id',
            full_name='suitebackend.Dataset.id',
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=_b("").decode('utf-8'),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='name',
            full_name='suitebackend.Dataset.name',
            index=1,
            number=2,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=_b("").decode('utf-8'),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='desc',
            full_name='suitebackend.Dataset.desc',
            index=2,
            number=3,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=_b("").decode('utf-8'),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='type',
            full_name='suitebackend.Dataset.type',
            index=3,
            number=4,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=_b("").decode('utf-8'),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='path',
            full_name='suitebackend.Dataset.path',
            index=4,
            number=5,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=_b("").decode('utf-8'),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='create_time',
            full_name='suitebackend.Dataset.create_time',
            index=5,
            number=6,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=_b("").decode('utf-8'),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax='proto3',
    extension_ranges=[],
    oneofs=[],
    serialized_start=33,
    serialized_end=131, )

_PROJECT = _descriptor.Descriptor(
    name='Project',
    full_name='suitebackend.Project',
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name='id',
            full_name='suitebackend.Project.id',
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=_b("").decode('utf-8'),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='name',
            full_name='suitebackend.Project.name',
            index=1,
            number=2,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=_b("").decode('utf-8'),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='desc',
            full_name='suitebackend.Project.desc',
            index=2,
            number=3,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=_b("").decode('utf-8'),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='type',
            full_name='suitebackend.Project.type',
            index=3,
            number=4,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=_b("").decode('utf-8'),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='did',
            full_name='suitebackend.Project.did',
            index=4,
            number=5,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=_b("").decode('utf-8'),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='path',
            full_name='suitebackend.Project.path',
            index=5,
            number=6,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=_b("").decode('utf-8'),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='create_time',
            full_name='suitebackend.Project.create_time',
            index=6,
            number=7,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=_b("").decode('utf-8'),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax='proto3',
    extension_ranges=[],
    oneofs=[],
    serialized_start=133,
    serialized_end=244, )

_TASK = _descriptor.Descriptor(
    name='Task',
    full_name='suitebackend.Task',
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name='id',
            full_name='suitebackend.Task.id',
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=_b("").decode('utf-8'),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='name',
            full_name='suitebackend.Task.name',
            index=1,
            number=2,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=_b("").decode('utf-8'),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='desc',
            full_name='suitebackend.Task.desc',
            index=2,
            number=3,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=_b("").decode('utf-8'),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='pid',
            full_name='suitebackend.Task.pid',
            index=3,
            number=4,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=_b("").decode('utf-8'),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='path',
            full_name='suitebackend.Task.path',
            index=4,
            number=5,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=_b("").decode('utf-8'),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='create_time',
            full_name='suitebackend.Task.create_time',
            index=5,
            number=6,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=_b("").decode('utf-8'),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='parent_id',
            full_name='suitebackend.Task.parent_id',
            index=6,
            number=7,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=_b("").decode('utf-8'),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax='proto3',
    extension_ranges=[],
    oneofs=[],
    serialized_start=246,
    serialized_end=359, )

_PRETRAINEDMODEL = _descriptor.Descriptor(
    name='PretrainedModel',
    full_name='suitebackend.PretrainedModel',
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name='id',
            full_name='suitebackend.PretrainedModel.id',
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=_b("").decode('utf-8'),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='name',
            full_name='suitebackend.PretrainedModel.name',
            index=1,
            number=2,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=_b("").decode('utf-8'),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='model',
            full_name='suitebackend.PretrainedModel.model',
            index=2,
            number=3,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=_b("").decode('utf-8'),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='type',
            full_name='suitebackend.PretrainedModel.type',
            index=3,
            number=4,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=_b("").decode('utf-8'),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='pid',
            full_name='suitebackend.PretrainedModel.pid',
            index=4,
            number=5,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=_b("").decode('utf-8'),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='tid',
            full_name='suitebackend.PretrainedModel.tid',
            index=5,
            number=6,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=_b("").decode('utf-8'),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='create_time',
            full_name='suitebackend.PretrainedModel.create_time',
            index=6,
            number=7,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=_b("").decode('utf-8'),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='path',
            full_name='suitebackend.PretrainedModel.path',
            index=7,
            number=8,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=_b("").decode('utf-8'),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax='proto3',
    extension_ranges=[],
    oneofs=[],
    serialized_start=362,
    serialized_end=495, )

_EXPORTEDMODEL = _descriptor.Descriptor(
    name='ExportedModel',
    full_name='suitebackend.ExportedModel',
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name='id',
            full_name='suitebackend.ExportedModel.id',
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=_b("").decode('utf-8'),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='name',
            full_name='suitebackend.ExportedModel.name',
            index=1,
            number=2,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=_b("").decode('utf-8'),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='model',
            full_name='suitebackend.ExportedModel.model',
            index=2,
            number=3,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=_b("").decode('utf-8'),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='type',
            full_name='suitebackend.ExportedModel.type',
            index=3,
            number=4,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=_b("").decode('utf-8'),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='pid',
            full_name='suitebackend.ExportedModel.pid',
            index=4,
            number=5,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=_b("").decode('utf-8'),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='tid',
            full_name='suitebackend.ExportedModel.tid',
            index=5,
            number=6,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=_b("").decode('utf-8'),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='create_time',
            full_name='suitebackend.ExportedModel.create_time',
            index=6,
            number=7,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=_b("").decode('utf-8'),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='path',
            full_name='suitebackend.ExportedModel.path',
            index=7,
            number=8,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=_b("").decode('utf-8'),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='exported_type',
            full_name='suitebackend.ExportedModel.exported_type',
            index=8,
            number=9,
            type=5,
            cpp_type=1,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax='proto3',
    extension_ranges=[],
    oneofs=[],
    serialized_start=498,
    serialized_end=652, )

_WORKSPACE_DATASETSENTRY = _descriptor.Descriptor(
    name='DatasetsEntry',
    full_name='suitebackend.Workspace.DatasetsEntry',
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name='key',
            full_name='suitebackend.Workspace.DatasetsEntry.key',
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=_b("").decode('utf-8'),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='value',
            full_name='suitebackend.Workspace.DatasetsEntry.value',
            index=1,
            number=2,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=_b('8\001'),
    is_extendable=False,
    syntax='proto3',
    extension_ranges=[],
    oneofs=[],
    serialized_start=1163,
    serialized_end=1233, )

_WORKSPACE_PROJECTSENTRY = _descriptor.Descriptor(
    name='ProjectsEntry',
    full_name='suitebackend.Workspace.ProjectsEntry',
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name='key',
            full_name='suitebackend.Workspace.ProjectsEntry.key',
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=_b("").decode('utf-8'),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='value',
            full_name='suitebackend.Workspace.ProjectsEntry.value',
            index=1,
            number=2,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=_b('8\001'),
    is_extendable=False,
    syntax='proto3',
    extension_ranges=[],
    oneofs=[],
    serialized_start=1235,
    serialized_end=1305, )

_WORKSPACE_TASKSENTRY = _descriptor.Descriptor(
    name='TasksEntry',
    full_name='suitebackend.Workspace.TasksEntry',
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name='key',
            full_name='suitebackend.Workspace.TasksEntry.key',
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=_b("").decode('utf-8'),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='value',
            full_name='suitebackend.Workspace.TasksEntry.value',
            index=1,
            number=2,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=_b('8\001'),
    is_extendable=False,
    syntax='proto3',
    extension_ranges=[],
    oneofs=[],
    serialized_start=1307,
    serialized_end=1371, )

_WORKSPACE_PRETRAINEDMODELSENTRY = _descriptor.Descriptor(
    name='PretrainedModelsEntry',
    full_name='suitebackend.Workspace.PretrainedModelsEntry',
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name='key',
            full_name='suitebackend.Workspace.PretrainedModelsEntry.key',
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=_b("").decode('utf-8'),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='value',
            full_name='suitebackend.Workspace.PretrainedModelsEntry.value',
            index=1,
            number=2,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=_b('8\001'),
    is_extendable=False,
    syntax='proto3',
    extension_ranges=[],
    oneofs=[],
    serialized_start=1373,
    serialized_end=1459, )

_WORKSPACE_EXPORTEDMODELSENTRY = _descriptor.Descriptor(
    name='ExportedModelsEntry',
    full_name='suitebackend.Workspace.ExportedModelsEntry',
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name='key',
            full_name='suitebackend.Workspace.ExportedModelsEntry.key',
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=_b("").decode('utf-8'),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='value',
            full_name='suitebackend.Workspace.ExportedModelsEntry.value',
            index=1,
            number=2,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=_b('8\001'),
    is_extendable=False,
    syntax='proto3',
    extension_ranges=[],
    oneofs=[],
    serialized_start=1461,
    serialized_end=1543, )

_WORKSPACE = _descriptor.Descriptor(
    name='Workspace',
    full_name='suitebackend.Workspace',
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name='version',
            full_name='suitebackend.Workspace.version',
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=_b("").decode('utf-8'),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='path',
            full_name='suitebackend.Workspace.path',
            index=1,
            number=2,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=_b("").decode('utf-8'),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='datasets',
            full_name='suitebackend.Workspace.datasets',
            index=2,
            number=3,
            type=11,
            cpp_type=10,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='projects',
            full_name='suitebackend.Workspace.projects',
            index=3,
            number=4,
            type=11,
            cpp_type=10,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='tasks',
            full_name='suitebackend.Workspace.tasks',
            index=4,
            number=5,
            type=11,
            cpp_type=10,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='max_dataset_id',
            full_name='suitebackend.Workspace.max_dataset_id',
            index=5,
            number=6,
            type=5,
            cpp_type=1,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='max_project_id',
            full_name='suitebackend.Workspace.max_project_id',
            index=6,
            number=7,
            type=5,
            cpp_type=1,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='max_task_id',
            full_name='suitebackend.Workspace.max_task_id',
            index=7,
            number=8,
            type=5,
            cpp_type=1,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='current_time',
            full_name='suitebackend.Workspace.current_time',
            index=8,
            number=9,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=_b("").decode('utf-8'),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='max_pretrained_model_id',
            full_name='suitebackend.Workspace.max_pretrained_model_id',
            index=9,
            number=10,
            type=5,
            cpp_type=1,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='pretrained_models',
            full_name='suitebackend.Workspace.pretrained_models',
            index=10,
            number=11,
            type=11,
            cpp_type=10,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='max_exported_model_id',
            full_name='suitebackend.Workspace.max_exported_model_id',
            index=11,
            number=12,
            type=5,
            cpp_type=1,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='exported_models',
            full_name='suitebackend.Workspace.exported_models',
            index=12,
            number=13,
            type=11,
            cpp_type=10,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR),
    ],
    extensions=[],
    nested_types=[
        _WORKSPACE_DATASETSENTRY,
        _WORKSPACE_PROJECTSENTRY,
        _WORKSPACE_TASKSENTRY,
        _WORKSPACE_PRETRAINEDMODELSENTRY,
        _WORKSPACE_EXPORTEDMODELSENTRY,
    ],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax='proto3',
    extension_ranges=[],
    oneofs=[],
    serialized_start=655,
    serialized_end=1543, )

_WORKSPACE_DATASETSENTRY.fields_by_name['value'].message_type = _DATASET
_WORKSPACE_DATASETSENTRY.containing_type = _WORKSPACE
_WORKSPACE_PROJECTSENTRY.fields_by_name['value'].message_type = _PROJECT
_WORKSPACE_PROJECTSENTRY.containing_type = _WORKSPACE
_WORKSPACE_TASKSENTRY.fields_by_name['value'].message_type = _TASK
_WORKSPACE_TASKSENTRY.containing_type = _WORKSPACE
_WORKSPACE_PRETRAINEDMODELSENTRY.fields_by_name[
    'value'].message_type = _PRETRAINEDMODEL
_WORKSPACE_PRETRAINEDMODELSENTRY.containing_type = _WORKSPACE
_WORKSPACE_EXPORTEDMODELSENTRY.fields_by_name[
    'value'].message_type = _EXPORTEDMODEL
_WORKSPACE_EXPORTEDMODELSENTRY.containing_type = _WORKSPACE
_WORKSPACE.fields_by_name['datasets'].message_type = _WORKSPACE_DATASETSENTRY
_WORKSPACE.fields_by_name['projects'].message_type = _WORKSPACE_PROJECTSENTRY
_WORKSPACE.fields_by_name['tasks'].message_type = _WORKSPACE_TASKSENTRY
_WORKSPACE.fields_by_name[
    'pretrained_models'].message_type = _WORKSPACE_PRETRAINEDMODELSENTRY
_WORKSPACE.fields_by_name[
    'exported_models'].message_type = _WORKSPACE_EXPORTEDMODELSENTRY
DESCRIPTOR.message_types_by_name['Dataset'] = _DATASET
DESCRIPTOR.message_types_by_name['Project'] = _PROJECT
DESCRIPTOR.message_types_by_name['Task'] = _TASK
DESCRIPTOR.message_types_by_name['PretrainedModel'] = _PRETRAINEDMODEL
DESCRIPTOR.message_types_by_name['ExportedModel'] = _EXPORTEDMODEL
DESCRIPTOR.message_types_by_name['Workspace'] = _WORKSPACE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Dataset = _reflection.GeneratedProtocolMessageType(
    'Dataset',
    (_message.Message, ),
    dict(
        DESCRIPTOR=_DATASET,
        __module__='workspace_pb2'
        # @@protoc_insertion_point(class_scope:suitebackend.Dataset)
    ))
_sym_db.RegisterMessage(Dataset)

Project = _reflection.GeneratedProtocolMessageType(
    'Project',
    (_message.Message, ),
    dict(
        DESCRIPTOR=_PROJECT,
        __module__='workspace_pb2'
        # @@protoc_insertion_point(class_scope:suitebackend.Project)
    ))
_sym_db.RegisterMessage(Project)

Task = _reflection.GeneratedProtocolMessageType(
    'Task',
    (_message.Message, ),
    dict(
        DESCRIPTOR=_TASK,
        __module__='workspace_pb2'
        # @@protoc_insertion_point(class_scope:suitebackend.Task)
    ))
_sym_db.RegisterMessage(Task)

PretrainedModel = _reflection.GeneratedProtocolMessageType(
    'PretrainedModel',
    (_message.Message, ),
    dict(
        DESCRIPTOR=_PRETRAINEDMODEL,
        __module__='workspace_pb2'
        # @@protoc_insertion_point(class_scope:suitebackend.PretrainedModel)
    ))
_sym_db.RegisterMessage(PretrainedModel)

ExportedModel = _reflection.GeneratedProtocolMessageType(
    'ExportedModel',
    (_message.Message, ),
    dict(
        DESCRIPTOR=_EXPORTEDMODEL,
        __module__='workspace_pb2'
        # @@protoc_insertion_point(class_scope:suitebackend.ExportedModel)
    ))
_sym_db.RegisterMessage(ExportedModel)

Workspace = _reflection.GeneratedProtocolMessageType(
    'Workspace',
    (_message.Message, ),
    dict(
        DatasetsEntry=_reflection.GeneratedProtocolMessageType(
            'DatasetsEntry',
            (_message.Message, ),
            dict(
                DESCRIPTOR=_WORKSPACE_DATASETSENTRY,
                __module__='workspace_pb2'
                # @@protoc_insertion_point(class_scope:suitebackend.Workspace.DatasetsEntry)
            )),
        ProjectsEntry=_reflection.GeneratedProtocolMessageType(
            'ProjectsEntry',
            (_message.Message, ),
            dict(
                DESCRIPTOR=_WORKSPACE_PROJECTSENTRY,
                __module__='workspace_pb2'
                # @@protoc_insertion_point(class_scope:suitebackend.Workspace.ProjectsEntry)
            )),
        TasksEntry=_reflection.GeneratedProtocolMessageType(
            'TasksEntry',
            (_message.Message, ),
            dict(
                DESCRIPTOR=_WORKSPACE_TASKSENTRY,
                __module__='workspace_pb2'
                # @@protoc_insertion_point(class_scope:suitebackend.Workspace.TasksEntry)
            )),
        PretrainedModelsEntry=_reflection.GeneratedProtocolMessageType(
            'PretrainedModelsEntry',
            (_message.Message, ),
            dict(
                DESCRIPTOR=_WORKSPACE_PRETRAINEDMODELSENTRY,
                __module__='workspace_pb2'
                # @@protoc_insertion_point(class_scope:suitebackend.Workspace.PretrainedModelsEntry)
            )),
        ExportedModelsEntry=_reflection.GeneratedProtocolMessageType(
            'ExportedModelsEntry',
            (_message.Message, ),
            dict(
                DESCRIPTOR=_WORKSPACE_EXPORTEDMODELSENTRY,
                __module__='workspace_pb2'
                # @@protoc_insertion_point(class_scope:suitebackend.Workspace.ExportedModelsEntry)
            )),
        DESCRIPTOR=_WORKSPACE,
        __module__='workspace_pb2'
        # @@protoc_insertion_point(class_scope:suitebackend.Workspace)
    ))
_sym_db.RegisterMessage(Workspace)
_sym_db.RegisterMessage(Workspace.DatasetsEntry)
_sym_db.RegisterMessage(Workspace.ProjectsEntry)
_sym_db.RegisterMessage(Workspace.TasksEntry)
_sym_db.RegisterMessage(Workspace.PretrainedModelsEntry)
_sym_db.RegisterMessage(Workspace.ExportedModelsEntry)

_WORKSPACE_DATASETSENTRY._options = None
_WORKSPACE_PROJECTSENTRY._options = None
_WORKSPACE_TASKSENTRY._options = None
_WORKSPACE_PRETRAINEDMODELSENTRY._options = None
_WORKSPACE_EXPORTEDMODELSENTRY._options = None
# @@protoc_insertion_point(module_scope)
