# copyright (c) 2020 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
from .cv.models.slim import prune
from .cv.models.slim import visualize

cal_params_sensitivities = prune.cal_params_sensitivities
visualize = visualize.visualize


def export_quant_model(model,
                       test_dataset,
                       batch_size=1,
                       batch_num=10,
                       save_dir='./quant_model',
                       cache_dir='./temp'):
    model.export_quant_model(
        dataset=test_dataset,
        batch_size=batch_size,
        batch_num=batch_num,
        save_dir=save_dir,
        cache_dir=cache_dir)
